@@ -1,4 +1,5 @@
 // Copyright 2021 Google LLC
+// SPDX-License-Identifier: Apache-2.0
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -33,7 +34,8 @@
 #include <vector>
 
 #undef VQSORT_TEST_IMPL
-#if (HWY_TARGET == HWY_SCALAR) || (defined(_MSC_VER) && !HWY_IS_DEBUG_BUILD)
+#if (HWY_TARGET == HWY_SCALAR || HWY_TARGET == HWY_EMU128) || \
+    (defined(_MSC_VER) && !HWY_IS_DEBUG_BUILD)
 // Scalar does not implement these, and MSVC non-debug builds time out.
 #define VQSORT_TEST_IMPL 0
 #else
@@ -54,7 +56,7 @@ namespace HWY_NAMESPACE {
 namespace {
 
 #if VQSORT_TEST_IMPL || VQSORT_TEST_SORT
-using detail::LaneTraits;
+using detail::TraitsLane;
 using detail::OrderAscending;
 using detail::OrderAscending128;
 using detail::OrderDescending;
@@ -89,7 +91,7 @@ static HWY_NOINLINE void TestMedian3() {
 }
 
 HWY_NOINLINE void TestAllMedian() {
-  TestMedian3<LaneTraits<OrderAscending> >();
+  TestMedian3<TraitsLane<OrderAscending> >();
 }
 
 template <class Traits, typename T>
@@ -234,8 +236,8 @@ HWY_NOINLINE void TestAllBaseCase() {
   return;
 #endif
 
-  TestBaseCase<LaneTraits<OrderAscending>, int32_t>();
-  TestBaseCase<LaneTraits<OrderDescending>, int64_t>();
+  TestBaseCase<TraitsLane<OrderAscending>, int32_t>();
+  TestBaseCase<TraitsLane<OrderDescending>, int64_t>();
   TestBaseCase<Traits128<OrderAscending128>, uint64_t>();
   TestBaseCase<Traits128<OrderDescending128>, uint64_t>();
 }
@@ -361,12 +363,12 @@ static HWY_NOINLINE void TestPartition() {
 }
 
 HWY_NOINLINE void TestAllPartition() {
-  TestPartition<LaneTraits<OrderAscending>, int16_t>();
-  TestPartition<LaneTraits<OrderDescending>, int32_t>();
-  TestPartition<LaneTraits<OrderAscending>, int64_t>();
-  TestPartition<LaneTraits<OrderDescending>, float>();
+  TestPartition<TraitsLane<OrderAscending>, int16_t>();
+  TestPartition<TraitsLane<OrderDescending>, int32_t>();
+  TestPartition<TraitsLane<OrderAscending>, int64_t>();
+  TestPartition<TraitsLane<OrderDescending>, float>();
 #if HWY_HAVE_FLOAT64
-  TestPartition<LaneTraits<OrderDescending>, double>();
+  TestPartition<TraitsLane<OrderDescending>, double>();
 #endif
   TestPartition<Traits128<OrderAscending128>, uint64_t>();
   TestPartition<Traits128<OrderDescending128>, uint64_t>();
@@ -479,6 +481,9 @@ void TestSort(size_t num) {
   SharedState shared;
   SharedTraits<Traits> st;
 
+  // Round up to a whole number of keys.
+  num += (st.Is128() && (num & 1));
+
   constexpr size_t kMaxMisalign = 16;
   auto aligned = hwy::AllocateAligned<T>(kMaxMisalign + num + kMaxMisalign);
   for (Algo algo : AlgoForTest()) {
@@ -531,24 +536,24 @@ void TestSort(size_t num) {
 }
 
 void TestAllSort() {
-  const size_t num = 15 * 1000;
+  const size_t num = AdjustedReps(20 * 1000);
 
-  TestSort<LaneTraits<OrderAscending>, int16_t>(num);
-  TestSort<LaneTraits<OrderDescending>, uint16_t>(num);
+  TestSort<TraitsLane<OrderAscending>, int16_t>(num);
+  TestSort<TraitsLane<OrderDescending>, uint16_t>(num);
 
-  TestSort<LaneTraits<OrderDescending>, int32_t>(num);
-  TestSort<LaneTraits<OrderDescending>, uint32_t>(num);
+  TestSort<TraitsLane<OrderDescending>, int32_t>(num);
+  TestSort<TraitsLane<OrderDescending>, uint32_t>(num);
 
-  TestSort<LaneTraits<OrderAscending>, int64_t>(num);
-  TestSort<LaneTraits<OrderAscending>, uint64_t>(num);
+  TestSort<TraitsLane<OrderAscending>, int64_t>(num);
+  TestSort<TraitsLane<OrderAscending>, uint64_t>(num);
 
   // WARNING: for float types, SIMD comparisons will flush denormals to zero,
   // causing mismatches with scalar sorts. In this test, we avoid generating
   // denormal inputs.
-  TestSort<LaneTraits<OrderAscending>, float>(num);
+  TestSort<TraitsLane<OrderAscending>, float>(num);
 #if HWY_HAVE_FLOAT64  // protects algo-inl's GenerateRandom
   if (Sorter::HaveFloat64()) {
-    TestSort<LaneTraits<OrderDescending>, double>(num);
+    TestSort<TraitsLane<OrderDescending>, double>(num);
   }
 #endif
 
@@ -577,10 +582,4 @@ HWY_EXPORT_AND_TEST_P(SortTest, TestAllSort);
 }  // namespace
 }  // namespace hwy
 
-// Ought not to be necessary, but without this, no tests run on RVV.
-int main(int argc, char** argv) {
-  ::testing::InitGoogleTest(&argc, argv);
-  return RUN_ALL_TESTS();
-}
-
 #endif  // HWY_ONCE
