@@ -1,4 +1,5 @@
 // Copyright 2021 Google LLC
+// SPDX-License-Identifier: Apache-2.0
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -15,14 +16,20 @@
 // ARM SVE[2] vectors (length not known at compile time).
 // External include guard in highway.h - see comment there.
 
+#include <arm_sve.h>
 #include <stddef.h>
 #include <stdint.h>
 
-#include <arm_sve.h>
-
 #include "hwy/base.h"
 #include "hwy/ops/shared-inl.h"
 
+// If running on hardware whose vector length is known to be a power of two, we
+// can skip fixups for non-power of two sizes. This may be 1 on future
+// fixed-size SVE targets.
+#ifndef HWY_SVE_IS_POW2
+#define HWY_SVE_IS_POW2 0
+#endif  // HWY_SVE_IS_POW2
+
 HWY_BEFORE_NAMESPACE();
 namespace hwy {
 namespace HWY_NAMESPACE {
@@ -357,8 +364,7 @@ HWY_API VFromD<D> BitCast(D d, FromV v) {
 // detail::*N() functions accept a scalar argument to avoid extra Set().
 
 // ------------------------------ Not
-
-HWY_SVE_FOREACH_UI(HWY_SVE_RETV_ARGPV, Not, not )
+HWY_SVE_FOREACH_UI(HWY_SVE_RETV_ARGPV, Not, not )  // NOLINT
 
 // ------------------------------ And
 
@@ -429,8 +435,13 @@ HWY_API V AndNot(const V a, const V b) {
   return BitCast(df, AndNot(BitCast(du, a), BitCast(du, b)));
 }
 
-// ------------------------------ OrAnd
+// ------------------------------ Or3
+template <class V>
+HWY_API V Or3(V o1, V o2, V o3) {
+  return Or(o1, Or(o2, o3));
+}
 
+// ------------------------------ OrAnd
 template <class V>
 HWY_API V OrAnd(const V o, const V a1, const V a2) {
   return Or(o, And(a1, a2));
@@ -601,6 +612,26 @@ HWY_SVE_FOREACH_UI32(HWY_SVE_RETV_ARGPVV, MulHigh, mulh)
 HWY_SVE_FOREACH_U64(HWY_SVE_RETV_ARGPVV, MulHigh, mulh)
 }  // namespace detail
 
+// ------------------------------ MulFixedPoint15
+HWY_API svint16_t MulFixedPoint15(svint16_t a, svint16_t b) {
+#if HWY_TARGET == HWY_SVE2
+  return svqrdmulh_s16(a, b);
+#else
+  const DFromV<decltype(a)> d;
+  const RebindToUnsigned<decltype(d)> du;
+
+  const svuint16_t lo = BitCast(du, Mul(a, b));
+  const svint16_t hi = MulHigh(a, b);
+  // We want (lo + 0x4000) >> 15, but that can overflow, and if it does we must
+  // carry that into the result. Instead isolate the top two bits because only
+  // they can influence the result.
+  const svuint16_t lo_top2 = ShiftRight<14>(lo);
+  // Bits 11: add 2, 10: add 1, 01: add 1, 00: add 0.
+  const svuint16_t rounding = ShiftRight<1>(detail::AddN(lo_top2, 1));
+  return Add(Add(hi, hi), BitCast(d, rounding));
+#endif
+}
+
 // ------------------------------ Div
 HWY_SVE_FOREACH_F(HWY_SVE_RETV_ARGPVV, Div, div)
 
@@ -774,7 +805,6 @@ HWY_SVE_FOREACH_F(HWY_SVE_COMPARE, Le, cmple)
 #undef HWY_SVE_COMPARE_N
 
 // ------------------------------ Gt/Ge (swapped order)
-
 template <class V>
 HWY_API svbool_t Gt(const V a, const V b) {
   return Lt(b, a);
@@ -817,6 +847,39 @@ HWY_API V IfVecThenElse(const V mask, const V yes, const V no) {
   return IfThenElse(MaskFromVec(mask), yes, no);
 }
 
+// ------------------------------ Floating-point classification (Ne)
+
+template <class V>
+HWY_API svbool_t IsNaN(const V v) {
+  return Ne(v, v);  // could also use cmpuo
+}
+
+template <class V>
+HWY_API svbool_t IsInf(const V v) {
+  using T = TFromV<V>;
+  const DFromV<decltype(v)> d;
+  const RebindToSigned<decltype(d)> di;
+  const VFromD<decltype(di)> vi = BitCast(di, v);
+  // 'Shift left' to clear the sign bit, check for exponent=max and mantissa=0.
+  return RebindMask(d, detail::EqN(Add(vi, vi), hwy::MaxExponentTimes2<T>()));
+}
+
+// Returns whether normal/subnormal/zero.
+template <class V>
+HWY_API svbool_t IsFinite(const V v) {
+  using T = TFromV<V>;
+  const DFromV<decltype(v)> d;
+  const RebindToUnsigned<decltype(d)> du;
+  const RebindToSigned<decltype(d)> di;  // cheaper than unsigned comparison
+  const VFromD<decltype(du)> vu = BitCast(du, v);
+  // 'Shift left' to clear the sign bit, then right so we can compare with the
+  // max exponent (cannot compare with MaxExponentTimes2 directly because it is
+  // negative and non-negative floats would be greater).
+  const VFromD<decltype(di)> exp =
+      BitCast(di, ShiftRight<hwy::MantissaBits<T>() + 1>(Add(vu, vu)));
+  return RebindMask(d, detail::LtN(exp, hwy::MaxExponentField<T>()));
+}
+
 // ================================================== MEMORY
 
 // ------------------------------ Load/MaskedLoad/LoadDup128/Store/Stream
@@ -854,12 +917,12 @@ HWY_API V IfVecThenElse(const V mask, const V yes, const V no) {
     sv##OP##_##CHAR##BITS(detail::MakeMask(d), p, v);         \
   }
 
-#define HWY_SVE_MASKED_STORE(BASE, CHAR, BITS, HALF, NAME, OP) \
-  template <size_t N, int kPow2>                               \
-  HWY_API void NAME(svbool_t m, HWY_SVE_V(BASE, BITS) v,       \
-                    HWY_SVE_D(BASE, BITS, N, kPow2) /* d */,   \
-                    HWY_SVE_T(BASE, BITS) * HWY_RESTRICT p) {  \
-    sv##OP##_##CHAR##BITS(m, p, v);                            \
+#define HWY_SVE_BLENDED_STORE(BASE, CHAR, BITS, HALF, NAME, OP) \
+  template <size_t N, int kPow2>                                \
+  HWY_API void NAME(HWY_SVE_V(BASE, BITS) v, svbool_t m,        \
+                    HWY_SVE_D(BASE, BITS, N, kPow2) /* d */,    \
+                    HWY_SVE_T(BASE, BITS) * HWY_RESTRICT p) {   \
+    sv##OP##_##CHAR##BITS(m, p, v);                             \
   }
 
 HWY_SVE_FOREACH(HWY_SVE_LOAD, Load, ld1)
@@ -867,13 +930,13 @@ HWY_SVE_FOREACH(HWY_SVE_MASKED_LOAD, MaskedLoad, ld1)
 HWY_SVE_FOREACH(HWY_SVE_LOAD_DUP128, LoadDup128, ld1rq)
 HWY_SVE_FOREACH(HWY_SVE_STORE, Store, st1)
 HWY_SVE_FOREACH(HWY_SVE_STORE, Stream, stnt1)
-HWY_SVE_FOREACH(HWY_SVE_MASKED_STORE, MaskedStore, st1)
+HWY_SVE_FOREACH(HWY_SVE_BLENDED_STORE, BlendedStore, st1)
 
 #undef HWY_SVE_LOAD
 #undef HWY_SVE_MASKED_LOAD
 #undef HWY_SVE_LOAD_DUP128
 #undef HWY_SVE_STORE
-#undef HWY_SVE_MASKED_STORE
+#undef HWY_SVE_BLENDED_STORE
 
 // BF16 is the same as svuint16_t because BF16 is optional before v8.6.
 template <size_t N, int kPow2>
@@ -955,6 +1018,80 @@ HWY_SVE_FOREACH_UIF3264(HWY_SVE_GATHER_INDEX, GatherIndex, ld1_gather)
 #undef HWY_SVE_GATHER_OFFSET
 #undef HWY_SVE_GATHER_INDEX
 
+// ------------------------------ LoadInterleaved2
+
+// Per-target flag to prevent generic_ops-inl.h from defining LoadInterleaved2.
+#ifdef HWY_NATIVE_LOAD_STORE_INTERLEAVED
+#undef HWY_NATIVE_LOAD_STORE_INTERLEAVED
+#else
+#define HWY_NATIVE_LOAD_STORE_INTERLEAVED
+#endif
+
+#define HWY_SVE_LOAD2(BASE, CHAR, BITS, HALF, NAME, OP)                       \
+  template <size_t N, int kPow2>                                              \
+  HWY_API void NAME(HWY_SVE_D(BASE, BITS, N, kPow2) d,                        \
+                    const HWY_SVE_T(BASE, BITS) * HWY_RESTRICT unaligned,     \
+                    HWY_SVE_V(BASE, BITS) & v0, HWY_SVE_V(BASE, BITS) & v1) { \
+    const sv##BASE##BITS##x2_t tuple =                                        \
+        sv##OP##_##CHAR##BITS(detail::MakeMask(d), unaligned);                \
+    v0 = svget2(tuple, 0);                                                    \
+    v1 = svget2(tuple, 1);                                                    \
+  }
+HWY_SVE_FOREACH(HWY_SVE_LOAD2, LoadInterleaved2, ld2)
+
+#undef HWY_SVE_LOAD2
+
+// ------------------------------ LoadInterleaved3
+
+#define HWY_SVE_LOAD3(BASE, CHAR, BITS, HALF, NAME, OP)                     \
+  template <size_t N, int kPow2>                                            \
+  HWY_API void NAME(HWY_SVE_D(BASE, BITS, N, kPow2) d,                      \
+                    const HWY_SVE_T(BASE, BITS) * HWY_RESTRICT unaligned,   \
+                    HWY_SVE_V(BASE, BITS) & v0, HWY_SVE_V(BASE, BITS) & v1, \
+                    HWY_SVE_V(BASE, BITS) & v2) {                           \
+    const sv##BASE##BITS##x3_t tuple =                                      \
+        sv##OP##_##CHAR##BITS(detail::MakeMask(d), unaligned);              \
+    v0 = svget3(tuple, 0);                                                  \
+    v1 = svget3(tuple, 1);                                                  \
+    v2 = svget3(tuple, 2);                                                  \
+  }
+HWY_SVE_FOREACH(HWY_SVE_LOAD3, LoadInterleaved3, ld3)
+
+#undef HWY_SVE_LOAD3
+
+// ------------------------------ LoadInterleaved4
+
+#define HWY_SVE_LOAD4(BASE, CHAR, BITS, HALF, NAME, OP)                       \
+  template <size_t N, int kPow2>                                              \
+  HWY_API void NAME(HWY_SVE_D(BASE, BITS, N, kPow2) d,                        \
+                    const HWY_SVE_T(BASE, BITS) * HWY_RESTRICT unaligned,     \
+                    HWY_SVE_V(BASE, BITS) & v0, HWY_SVE_V(BASE, BITS) & v1,   \
+                    HWY_SVE_V(BASE, BITS) & v2, HWY_SVE_V(BASE, BITS) & v3) { \
+    const sv##BASE##BITS##x4_t tuple =                                        \
+        sv##OP##_##CHAR##BITS(detail::MakeMask(d), unaligned);                \
+    v0 = svget4(tuple, 0);                                                    \
+    v1 = svget4(tuple, 1);                                                    \
+    v2 = svget4(tuple, 2);                                                    \
+    v3 = svget4(tuple, 3);                                                    \
+  }
+HWY_SVE_FOREACH(HWY_SVE_LOAD4, LoadInterleaved4, ld4)
+
+#undef HWY_SVE_LOAD4
+
+// ------------------------------ StoreInterleaved2
+
+#define HWY_SVE_STORE2(BASE, CHAR, BITS, HALF, NAME, OP)                 \
+  template <size_t N, int kPow2>                                         \
+  HWY_API void NAME(HWY_SVE_V(BASE, BITS) v0, HWY_SVE_V(BASE, BITS) v1,  \
+                    HWY_SVE_D(BASE, BITS, N, kPow2) d,                   \
+                    HWY_SVE_T(BASE, BITS) * HWY_RESTRICT unaligned) {    \
+    const sv##BASE##BITS##x2_t tuple = svcreate2##_##CHAR##BITS(v0, v1); \
+    sv##OP##_##CHAR##BITS(detail::MakeMask(d), unaligned, tuple);        \
+  }
+HWY_SVE_FOREACH(HWY_SVE_STORE2, StoreInterleaved2, st2)
+
+#undef HWY_SVE_STORE2
+
 // ------------------------------ StoreInterleaved3
 
 #define HWY_SVE_STORE3(BASE, CHAR, BITS, HALF, NAME, OP)                      \
@@ -966,7 +1103,7 @@ HWY_SVE_FOREACH_UIF3264(HWY_SVE_GATHER_INDEX, GatherIndex, ld1_gather)
     const sv##BASE##BITS##x3_t triple = svcreate3##_##CHAR##BITS(v0, v1, v2); \
     sv##OP##_##CHAR##BITS(detail::MakeMask(d), unaligned, triple);            \
   }
-HWY_SVE_FOREACH_U08(HWY_SVE_STORE3, StoreInterleaved3, st3)
+HWY_SVE_FOREACH(HWY_SVE_STORE3, StoreInterleaved3, st3)
 
 #undef HWY_SVE_STORE3
 
@@ -982,7 +1119,7 @@ HWY_SVE_FOREACH_U08(HWY_SVE_STORE3, StoreInterleaved3, st3)
         svcreate4##_##CHAR##BITS(v0, v1, v2, v3);                       \
     sv##OP##_##CHAR##BITS(detail::MakeMask(d), unaligned, quad);        \
   }
-HWY_SVE_FOREACH_U08(HWY_SVE_STORE4, StoreInterleaved4, st4)
+HWY_SVE_FOREACH(HWY_SVE_STORE4, StoreInterleaved4, st4)
 
 #undef HWY_SVE_STORE4
 
@@ -1013,10 +1150,6 @@ HWY_API svint32_t PromoteTo(Simd<int32_t, N, kPow2> dto, svint8_t vfrom) {
   const RepartitionToWide<DFromV<decltype(vfrom)>> d2;
   return PromoteTo(dto, PromoteTo(d2, vfrom));
 }
-template <size_t N, int kPow2>
-HWY_API svuint32_t U32FromU8(svuint8_t v) {
-  return PromoteTo(Simd<uint32_t, N, kPow2>(), v);
-}
 
 // Sign change
 template <size_t N, int kPow2>
@@ -1214,7 +1347,8 @@ HWY_SVE_FOREACH(HWY_SVE_SPLICE, Splice, splice)
 
 template <class D>
 HWY_API VFromD<D> ConcatOdd(D d, VFromD<D> hi, VFromD<D> lo) {
-#if 0  // if we could assume VL is a power of two
+#if HWY_SVE_IS_POW2
+  (void)d;
   return detail::ConcatOdd(hi, lo);
 #else
   const VFromD<D> hi_odd = detail::ConcatOdd(hi, hi);
@@ -1225,7 +1359,8 @@ HWY_API VFromD<D> ConcatOdd(D d, VFromD<D> hi, VFromD<D> lo) {
 
 template <class D>
 HWY_API VFromD<D> ConcatEven(D d, VFromD<D> hi, VFromD<D> lo) {
-#if 0  // if we could assume VL is a power of two
+#if HWY_SVE_IS_POW2
+  (void)d;
   return detail::ConcatEven(hi, lo);
 #else
   const VFromD<D> hi_odd = detail::ConcatEven(hi, hi);
@@ -1280,7 +1415,6 @@ HWY_SVE_FOREACH_F(HWY_SVE_CONVERT, ConvertTo, cvt)
 #undef HWY_SVE_CONVERT
 
 // ------------------------------ NearestInt (Round, ConvertTo)
-
 template <class VF, class DI = RebindToSigned<DFromV<VF>>>
 HWY_API VFromD<DI> NearestInt(VF v) {
   // No single instruction, round then truncate.
@@ -1365,7 +1499,6 @@ HWY_API VFromD<D> Combine(const D d, const V2 hi, const V2 lo) {
 }
 
 // ------------------------------ ZeroExtendVector
-
 template <class D, class V>
 HWY_API V ZeroExtendVector(const D d, const V lo) {
   return Combine(d, Zero(Half<D>()), lo);
@@ -1392,13 +1525,35 @@ HWY_API V UpperHalf(const D2 /* d2 */, const V v) {
 
 // ------------------------------ GetLane
 
-#define HWY_SVE_GET_LANE(BASE, CHAR, BITS, HALF, NAME, OP)      \
-  HWY_API HWY_SVE_T(BASE, BITS) NAME(HWY_SVE_V(BASE, BITS) v) { \
-    return sv##OP##_##CHAR##BITS(detail::PFalse(), v);          \
+namespace detail {
+#define HWY_SVE_GET_LANE(BASE, CHAR, BITS, HALF, NAME, OP) \
+  HWY_INLINE HWY_SVE_T(BASE, BITS)                         \
+      NAME(HWY_SVE_V(BASE, BITS) v, svbool_t mask) {       \
+    return sv##OP##_##CHAR##BITS(mask, v);                 \
   }
 
 HWY_SVE_FOREACH(HWY_SVE_GET_LANE, GetLane, lasta)
 #undef HWY_SVE_GET_LANE
+}  // namespace detail
+
+template <class V>
+HWY_API TFromV<V> GetLane(V v) {
+  return detail::GetLane(v, detail::PFalse());
+}
+
+// ------------------------------ ExtractLane
+template <class V>
+HWY_API TFromV<V> ExtractLane(V v, size_t i) {
+  return detail::GetLane(v, FirstN(DFromV<V>(), i));
+}
+
+// ------------------------------ InsertLane (IfThenElse)
+template <class V>
+HWY_API V InsertLane(const V v, size_t i, TFromV<V> t) {
+  const DFromV<V> d;
+  const auto is_i = detail::EqN(Iota(d, 0), static_cast<TFromV<V>>(i));
+  return IfThenElse(RebindMask(d, is_i), Set(d, t), v);
+}
 
 // ------------------------------ DupEven
 
@@ -1502,27 +1657,32 @@ HWY_API V SwapAdjacentBlocks(const V v) {
 
 // ------------------------------ Reverse
 
-#if 0  // if we could assume VL is a power of two
-#error "Update macro"
-#endif
-#define HWY_SVE_REVERSE(BASE, CHAR, BITS, HALF, NAME, OP)                \
-  template <size_t N, int kPow2>                                         \
-  HWY_API HWY_SVE_V(BASE, BITS)                                          \
-      NAME(HWY_SVE_D(BASE, BITS, N, kPow2) d, HWY_SVE_V(BASE, BITS) v) { \
-    const auto reversed = sv##OP##_##CHAR##BITS(v);                      \
-    /* Shift right to remove extra (non-pow2 and remainder) lanes. */    \
-    const size_t all_lanes =                                             \
-        detail::AllHardwareLanes(hwy::SizeTag<BITS / 8>());              \
-    /* TODO(janwas): on SVE2, use whilege. */                            \
-    /* Avoids FirstN truncating to the return vector size. */            \
-    const ScalableTag<HWY_SVE_T(BASE, BITS)> dfull;                      \
-    const svbool_t mask = Not(FirstN(dfull, all_lanes - Lanes(d)));      \
-    return detail::Splice(reversed, reversed, mask);                     \
+namespace detail {
+
+#define HWY_SVE_REVERSE(BASE, CHAR, BITS, HALF, NAME, OP)       \
+  HWY_API HWY_SVE_V(BASE, BITS) NAME(HWY_SVE_V(BASE, BITS) v) { \
+    return sv##OP##_##CHAR##BITS(v);                            \
   }
 
-HWY_SVE_FOREACH(HWY_SVE_REVERSE, Reverse, rev)
+HWY_SVE_FOREACH(HWY_SVE_REVERSE, ReverseFull, rev)
 #undef HWY_SVE_REVERSE
 
+}  // namespace detail
+
+template <class D, class V>
+HWY_API V Reverse(D d, V v) {
+  using T = TFromD<D>;
+  const auto reversed = detail::ReverseFull(v);
+  if (HWY_SVE_IS_POW2 && detail::IsFull(d)) return reversed;
+  // Shift right to remove extra (non-pow2 and remainder) lanes.
+  // TODO(janwas): on SVE2, use whilege.
+  const size_t all_lanes = detail::AllHardwareLanes(hwy::SizeTag<sizeof(T)>());
+  // Avoids FirstN truncating to the return vector size.
+  const ScalableTag<T> dfull;
+  const svbool_t mask = Not(FirstN(dfull, all_lanes - Lanes(d)));
+  return detail::Splice(reversed, reversed, mask);
+}
+
 // ------------------------------ Reverse2
 
 template <class D, HWY_IF_LANE_SIZE_D(D, 2)>
@@ -1546,17 +1706,15 @@ HWY_API VFromD<D> Reverse2(D /* tag */, const VFromD<D> v) {  // 3210
 }
 
 // ------------------------------ Reverse4 (TableLookupLanes)
-
-// TODO(janwas): is this approach faster than Shuffle0123?
 template <class D>
 HWY_API VFromD<D> Reverse4(D d, const VFromD<D> v) {
+  // TODO(janwas): is this approach faster than Shuffle0123?
   const RebindToUnsigned<decltype(d)> du;
   const auto idx = detail::XorN(Iota(du, 0), 3);
   return TableLookupLanes(v, idx);
 }
 
 // ------------------------------ Reverse8 (TableLookupLanes)
-
 template <class D>
 HWY_API VFromD<D> Reverse8(D d, const VFromD<D> v) {
   const RebindToUnsigned<decltype(d)> du;
@@ -1566,6 +1724,11 @@ HWY_API VFromD<D> Reverse8(D d, const VFromD<D> v) {
 
 // ------------------------------ Compress (PromoteTo)
 
+template <typename T>
+struct CompressIsPartition {
+  enum { value = 0 };
+};
+
 #define HWY_SVE_COMPRESS(BASE, CHAR, BITS, HALF, NAME, OP)                     \
   HWY_API HWY_SVE_V(BASE, BITS) NAME(HWY_SVE_V(BASE, BITS) v, svbool_t mask) { \
     return sv##OP##_##CHAR##BITS(mask, v);                                     \
@@ -1611,7 +1774,6 @@ HWY_API svfloat16_t Compress(svfloat16_t v, svbool_t mask16) {
 }
 
 // ------------------------------ CompressStore
-
 template <class V, class M, class D>
 HWY_API size_t CompressStore(const V v, const M mask, const D d,
                              TFromD<D>* HWY_RESTRICT unaligned) {
@@ -1620,13 +1782,12 @@ HWY_API size_t CompressStore(const V v, const M mask, const D d,
 }
 
 // ------------------------------ CompressBlendedStore
-
 template <class V, class M, class D>
 HWY_API size_t CompressBlendedStore(const V v, const M mask, const D d,
                                     TFromD<D>* HWY_RESTRICT unaligned) {
   const size_t count = CountTrue(d, mask);
   const svbool_t store_mask = FirstN(d, count);
-  MaskedStore(store_mask, Compress(v, mask), d, unaligned);
+  BlendedStore(Compress(v, mask), store_mask, d, unaligned);
   return count;
 }
 
@@ -1644,12 +1805,45 @@ HWY_INLINE V OffsetsOf128BitBlocks(const D d, const V iota0) {
   return detail::AndNotN(static_cast<T>(LanesPerBlock(d) - 1), iota0);
 }
 
-template <size_t kLanes, class D>
+template <size_t kLanes, class D, HWY_IF_LANE_SIZE_D(D, 1)>
 svbool_t FirstNPerBlock(D d) {
-  const RebindToSigned<decltype(d)> di;
-  constexpr size_t kLanesPerBlock = detail::LanesPerBlock(di);
-  const auto idx_mod = detail::AndN(Iota(di, 0), kLanesPerBlock - 1);
-  return detail::LtN(BitCast(di, idx_mod), kLanes);
+  const RebindToUnsigned<decltype(d)> du;
+  constexpr size_t kLanesPerBlock = detail::LanesPerBlock(du);
+  const svuint8_t idx_mod =
+      svdupq_n_u8(0 % kLanesPerBlock, 1 % kLanesPerBlock, 2 % kLanesPerBlock,
+                  3 % kLanesPerBlock, 4 % kLanesPerBlock, 5 % kLanesPerBlock,
+                  6 % kLanesPerBlock, 7 % kLanesPerBlock, 8 % kLanesPerBlock,
+                  9 % kLanesPerBlock, 10 % kLanesPerBlock, 11 % kLanesPerBlock,
+                  12 % kLanesPerBlock, 13 % kLanesPerBlock, 14 % kLanesPerBlock,
+                  15 % kLanesPerBlock);
+  return detail::LtN(BitCast(du, idx_mod), kLanes);
+}
+template <size_t kLanes, class D, HWY_IF_LANE_SIZE_D(D, 2)>
+svbool_t FirstNPerBlock(D d) {
+  const RebindToUnsigned<decltype(d)> du;
+  constexpr size_t kLanesPerBlock = detail::LanesPerBlock(du);
+  const svuint16_t idx_mod =
+      svdupq_n_u16(0 % kLanesPerBlock, 1 % kLanesPerBlock, 2 % kLanesPerBlock,
+                   3 % kLanesPerBlock, 4 % kLanesPerBlock, 5 % kLanesPerBlock,
+                   6 % kLanesPerBlock, 7 % kLanesPerBlock);
+  return detail::LtN(BitCast(du, idx_mod), kLanes);
+}
+template <size_t kLanes, class D, HWY_IF_LANE_SIZE_D(D, 4)>
+svbool_t FirstNPerBlock(D d) {
+  const RebindToUnsigned<decltype(d)> du;
+  constexpr size_t kLanesPerBlock = detail::LanesPerBlock(du);
+  const svuint32_t idx_mod =
+      svdupq_n_u32(0 % kLanesPerBlock, 1 % kLanesPerBlock, 2 % kLanesPerBlock,
+                   3 % kLanesPerBlock);
+  return detail::LtN(BitCast(du, idx_mod), kLanes);
+}
+template <size_t kLanes, class D, HWY_IF_LANE_SIZE_D(D, 8)>
+svbool_t FirstNPerBlock(D d) {
+  const RebindToUnsigned<decltype(d)> du;
+  constexpr size_t kLanesPerBlock = detail::LanesPerBlock(du);
+  const svuint64_t idx_mod =
+      svdupq_n_u64(0 % kLanesPerBlock, 1 % kLanesPerBlock);
+  return detail::LtN(BitCast(du, idx_mod), kLanes);
 }
 
 }  // namespace detail
@@ -1666,7 +1860,6 @@ HWY_API V CombineShiftRightBytes(const D d, const V hi, const V lo) {
 }
 
 // ------------------------------ Shuffle2301
-
 template <class V>
 HWY_API V Shuffle2301(const V v) {
   const DFromV<V> d;
@@ -1752,7 +1945,6 @@ HWY_API VI TableLookupBytesOr0(const V v, const VI idx) {
 }
 
 // ------------------------------ Broadcast
-
 template <int kLane, class V>
 HWY_API V Broadcast(const V v) {
   const DFromV<V> d;
@@ -1906,7 +2098,6 @@ HWY_SVE_FOREACH_F(HWY_SVE_REDUCE, MaxOfLanes, maxnmv)
 // ================================================== Ops with dependencies
 
 // ------------------------------ PromoteTo bfloat16 (ZipLower)
-
 template <size_t N, int kPow2>
 HWY_API svfloat32_t PromoteTo(Simd<float32_t, N, kPow2> df32,
                               const svuint16_t v) {
@@ -1914,7 +2105,6 @@ HWY_API svfloat32_t PromoteTo(Simd<float32_t, N, kPow2> df32,
 }
 
 // ------------------------------ ReorderDemote2To (OddEven)
-
 template <size_t N, int kPow2>
 HWY_API svuint16_t ReorderDemote2To(Simd<bfloat16_t, N, kPow2> dbf16,
                                     svfloat32_t a, svfloat32_t b) {
@@ -1972,9 +2162,8 @@ HWY_INLINE svbool_t LoadMaskBits(D d, const uint8_t* HWY_RESTRICT bits) {
   // Replicate bytes 8x such that each byte contains the bit that governs it.
   const svuint8_t rep8 = svtbl_u8(bytes, detail::AndNotN(7, iota));
 
-  // 1, 2, 4, 8, 16, 32, 64, 128,  1, 2 ..
-  const svuint8_t bit = Shl(Set(du, 1), detail::AndN(iota, 7));
-
+  const svuint8_t bit =
+      svdupq_n_u8(1, 2, 4, 8, 16, 32, 64, 128, 1, 2, 4, 8, 16, 32, 64, 128);
   return TestBit(rep8, bit);
 }
 
@@ -1990,9 +2179,7 @@ HWY_INLINE svbool_t LoadMaskBits(D /* tag */,
   // Replicate bytes 16x such that each lane contains the bit that governs it.
   const svuint8_t rep16 = svtbl_u8(bytes, ShiftRight<4>(Iota(du8, 0)));
 
-  // 1, 2, 4, 8, 16, 32, 64, 128,  1, 2 ..
-  const svuint16_t bit = Shl(Set(du, 1), detail::AndN(Iota(du, 0), 7));
-
+  const svuint16_t bit = svdupq_n_u16(1, 2, 4, 8, 16, 32, 64, 128);
   return TestBit(BitCast(du, rep16), bit);
 }
 
@@ -2096,13 +2283,13 @@ HWY_API size_t StoreMaskBits(D d, svbool_t m, uint8_t* bits) {
   return num_bytes;
 }
 
-// ------------------------------ CompressBits, CompressBitsStore (LoadMaskBits)
-
+// ------------------------------ CompressBits (LoadMaskBits)
 template <class V>
 HWY_INLINE V CompressBits(V v, const uint8_t* HWY_RESTRICT bits) {
   return Compress(v, LoadMaskBits(DFromV<V>(), bits));
 }
 
+// ------------------------------ CompressBitsStore (LoadMaskBits)
 template <class D>
 HWY_API size_t CompressBitsStore(VFromD<D> v, const uint8_t* HWY_RESTRICT bits,
                                  D d, TFromD<D>* HWY_RESTRICT unaligned) {
@@ -2148,7 +2335,6 @@ HWY_API svuint64_t MulOdd(const svuint64_t a, const svuint64_t b) {
 }
 
 // ------------------------------ ReorderWidenMulAccumulate (MulAdd, ZipLower)
-
 template <size_t N, int kPow2>
 HWY_API svfloat32_t ReorderWidenMulAccumulate(Simd<float, N, kPow2> df32,
                                               svuint16_t a, svuint16_t b,
@@ -2198,7 +2384,6 @@ HWY_API svuint64_t CLMulUpper(const svuint64_t a, const svuint64_t b) {
 #endif  // __ARM_FEATURE_SVE2_AES
 
 // ------------------------------ Lt128
-
 template <class D>
 HWY_INLINE svbool_t Lt128(D /* d */, const svuint64_t a, const svuint64_t b) {
   static_assert(!IsSigned<TFromD<D>>() && sizeof(TFromD<D>) == 8, "Use u64");
@@ -2219,8 +2404,8 @@ HWY_INLINE svbool_t Lt128(D /* d */, const svuint64_t a, const svuint64_t b) {
   const svbool_t ltHL = Lt(a, b);
   // trn (interleave even/odd) allow us to move and copy masks across lanes.
   const svbool_t cmpLL = svtrn1_b64(ltHL, ltHL);
-  const svbool_t outHx = svsel_b(eqHL, cmpLL, ltHL);   // See truth table above.
-  return svtrn2_b64(outHx, outHx);                     // replicate to HH
+  const svbool_t outHx = svsel_b(eqHL, cmpLL, ltHL);  // See truth table above.
+  return svtrn2_b64(outHx, outHx);                    // replicate to HH
 }
 
 // ------------------------------ Min128, Max128 (Lt128)
