@@ -1,4 +1,5 @@
 // Copyright 2019 Google LLC
+// SPDX-License-Identifier: Apache-2.0
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -126,6 +127,7 @@ enum class FeatureIndex : uint32_t {
 
   kVNNI,
   kVPCLMULQDQ,
+  kVBMI,
   kVBMI2,
   kVAES,
   kPOPCNTDQ,
@@ -176,14 +178,15 @@ constexpr uint64_t kGroupAVX3 =
 
 constexpr uint64_t kGroupAVX3_DL =
     Bit(FeatureIndex::kVNNI) | Bit(FeatureIndex::kVPCLMULQDQ) |
-    Bit(FeatureIndex::kVBMI2) | Bit(FeatureIndex::kVAES) |
-    Bit(FeatureIndex::kPOPCNTDQ) | Bit(FeatureIndex::kBITALG) | kGroupAVX3;
+    Bit(FeatureIndex::kVBMI) | Bit(FeatureIndex::kVBMI2) |
+    Bit(FeatureIndex::kVAES) | Bit(FeatureIndex::kPOPCNTDQ) |
+    Bit(FeatureIndex::kBITALG) | kGroupAVX3;
 
 #endif  // HWY_ARCH_X86
 
 }  // namespace
 
-HWY_NORETURN void HWY_FORMAT(3, 4)
+HWY_DLLEXPORT HWY_NORETURN void HWY_FORMAT(3, 4)
     Abort(const char* file, int line, const char* format, ...) {
   char buf[2000];
   va_list args;
@@ -211,15 +214,16 @@ HWY_NORETURN void HWY_FORMAT(3, 4)
 #endif
 }
 
-void DisableTargets(uint32_t disabled_targets) {
-  supported_mask_ = ~(disabled_targets & ~uint32_t(HWY_ENABLED_BASELINE));
+HWY_DLLEXPORT void DisableTargets(uint32_t disabled_targets) {
+  supported_mask_ =
+      ~(disabled_targets & ~static_cast<uint32_t>(HWY_ENABLED_BASELINE));
   // We can call Update() here to initialize the mask but that will trigger a
   // call to SupportedTargets() which we use in tests to tell whether any of the
   // highway dynamic dispatch functions were used.
   GetChosenTarget().DeInit();
 }
 
-void SetSupportedTargetsForTest(uint32_t targets) {
+HWY_DLLEXPORT void SetSupportedTargetsForTest(uint32_t targets) {
   // Reset the cached supported_ value to 0 to force a re-evaluation in the
   // next call to SupportedTargets() which will use the mocked value set here
   // if not zero.
@@ -228,11 +232,11 @@ void SetSupportedTargetsForTest(uint32_t targets) {
   GetChosenTarget().DeInit();
 }
 
-bool SupportedTargetsCalledForTest() {
+HWY_DLLEXPORT bool SupportedTargetsCalledForTest() {
   return supported_.load(std::memory_order_acquire) != 0;
 }
 
-uint32_t SupportedTargets() {
+HWY_DLLEXPORT uint32_t SupportedTargets() {
   uint32_t bits = supported_.load(std::memory_order_acquire);
   // Already initialized?
   if (HWY_LIKELY(bits != 0)) {
@@ -246,7 +250,11 @@ uint32_t SupportedTargets() {
     return supported_targets_for_test_ & supported_mask_;
   }
 
+#if defined(HWY_COMPILE_ONLY_SCALAR)
   bits = HWY_SCALAR;
+#else
+  bits = HWY_EMU128;
+#endif
 
 #if HWY_ARCH_X86
   bool has_osxsave = false;
@@ -288,6 +296,7 @@ uint32_t SupportedTargets() {
       flags |= IsBitSet(abcd[1], 30) ? Bit(FeatureIndex::kAVX512BW) : 0;
       flags |= IsBitSet(abcd[1], 31) ? Bit(FeatureIndex::kAVX512VL) : 0;
 
+      flags |= IsBitSet(abcd[2], 1) ? Bit(FeatureIndex::kVBMI) : 0;
       flags |= IsBitSet(abcd[2], 6) ? Bit(FeatureIndex::kVBMI2) : 0;
       flags |= IsBitSet(abcd[2], 9) ? Bit(FeatureIndex::kVAES) : 0;
       flags |= IsBitSet(abcd[2], 10) ? Bit(FeatureIndex::kVPCLMULQDQ) : 0;
@@ -318,30 +327,33 @@ uint32_t SupportedTargets() {
   // are not preserved across context switches.
   if (has_osxsave) {
     const uint32_t xcr0 = ReadXCR0();
+    const uint32_t min_avx3 = HWY_AVX3 | HWY_AVX3_DL;
+    const uint32_t min_avx2 = HWY_AVX2 | min_avx3;
     // XMM
     if (!IsBitSet(xcr0, 1)) {
-      bits &=
-          ~uint32_t(HWY_SSSE3 | HWY_SSE4 | HWY_AVX2 | HWY_AVX3 | HWY_AVX3_DL);
+      bits &= ~(HWY_SSSE3 | HWY_SSE4 | min_avx2);
     }
     // YMM
     if (!IsBitSet(xcr0, 2)) {
-      bits &= ~uint32_t(HWY_AVX2 | HWY_AVX3 | HWY_AVX3_DL);
+      bits &= ~min_avx2;
     }
-    // ZMM + opmask
-    if ((xcr0 & 0x70) != 0x70) {
-      bits &= ~uint32_t(HWY_AVX3 | HWY_AVX3_DL);
+    // opmask, ZMM lo/hi
+    if (!IsBitSet(xcr0, 5) || !IsBitSet(xcr0, 6) || !IsBitSet(xcr0, 7)) {
+      bits &= ~min_avx3;
     }
   }
 
-#else
-  // TODO(janwas): detect for other platforms
-  bits = HWY_ENABLED_BASELINE;
-#endif  // HWY_ARCH_X86
-
   if ((bits & HWY_ENABLED_BASELINE) != HWY_ENABLED_BASELINE) {
     fprintf(stderr, "WARNING: CPU supports %zx but software requires %x\n",
             size_t(bits), HWY_ENABLED_BASELINE);
   }
+#else
+  // TODO(janwas): detect for other platforms and check for baseline
+  // This file is typically compiled without HWY_IS_TEST, but targets_test has
+  // it set, and will expect all of its HWY_TARGETS (= all attainable) to be
+  // supported.
+  bits |= HWY_ENABLED_BASELINE;
+#endif  // HWY_ARCH_X86
 
   supported_.store(bits, std::memory_order_release);
   return bits & supported_mask_;
@@ -352,14 +364,14 @@ HWY_DLLEXPORT ChosenTarget& GetChosenTarget() {
   return chosen_target;
 }
 
-void ChosenTarget::Update() {
+HWY_DLLEXPORT void ChosenTarget::Update() {
   // The supported variable contains the current CPU supported targets shifted
   // to the location expected by the ChosenTarget mask. We enabled SCALAR
   // regardless of whether it was compiled since it is also used as the
   // fallback mechanism to the baseline target.
   uint32_t supported = HWY_CHOSEN_TARGET_SHIFT(hwy::SupportedTargets()) |
                        HWY_CHOSEN_TARGET_MASK_SCALAR;
-  mask_.store(supported);
+  StoreMask(supported);
 }
 
 }  // namespace hwy
