@@ -1,4 +1,5 @@
 // Copyright 2019 Google LLC
+// SPDX-License-Identifier: Apache-2.0
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -26,101 +27,6 @@ HWY_BEFORE_NAMESPACE();
 namespace hwy {
 namespace HWY_NAMESPACE {
 
-struct TestShiftBytes {
-  template <class T, class D>
-  HWY_NOINLINE void operator()(T /*unused*/, D d) {
-    // Scalar does not define Shift*Bytes.
-#if HWY_TARGET != HWY_SCALAR || HWY_IDE
-    const Repartition<uint8_t, D> du8;
-    const size_t N8 = Lanes(du8);
-
-    // Zero remains zero
-    const auto v0 = Zero(d);
-    HWY_ASSERT_VEC_EQ(d, v0, ShiftLeftBytes<1>(v0));
-    HWY_ASSERT_VEC_EQ(d, v0, ShiftLeftBytes<1>(d, v0));
-    HWY_ASSERT_VEC_EQ(d, v0, ShiftRightBytes<1>(d, v0));
-
-    // Zero after shifting out the high/low byte
-    auto bytes = AllocateAligned<uint8_t>(N8);
-    std::fill(bytes.get(), bytes.get() + N8, 0);
-    bytes[N8 - 1] = 0x7F;
-    const auto vhi = BitCast(d, Load(du8, bytes.get()));
-    bytes[N8 - 1] = 0;
-    bytes[0] = 0x7F;
-    const auto vlo = BitCast(d, Load(du8, bytes.get()));
-    HWY_ASSERT_VEC_EQ(d, v0, ShiftLeftBytes<1>(vhi));
-    HWY_ASSERT_VEC_EQ(d, v0, ShiftLeftBytes<1>(d, vhi));
-    HWY_ASSERT_VEC_EQ(d, v0, ShiftRightBytes<1>(d, vlo));
-
-    // Check expected result with Iota
-    const size_t N = Lanes(d);
-    auto in = AllocateAligned<T>(N);
-    const uint8_t* in_bytes = reinterpret_cast<const uint8_t*>(in.get());
-    const auto v = BitCast(d, Iota(du8, 1));
-    Store(v, d, in.get());
-
-    auto expected = AllocateAligned<T>(N);
-    uint8_t* expected_bytes = reinterpret_cast<uint8_t*>(expected.get());
-
-    const size_t kBlockSize = HWY_MIN(N8, 16);
-    for (size_t block = 0; block < N8; block += kBlockSize) {
-      expected_bytes[block] = 0;
-      memcpy(expected_bytes + block + 1, in_bytes + block, kBlockSize - 1);
-    }
-    HWY_ASSERT_VEC_EQ(d, expected.get(), ShiftLeftBytes<1>(v));
-    HWY_ASSERT_VEC_EQ(d, expected.get(), ShiftLeftBytes<1>(d, v));
-
-    for (size_t block = 0; block < N8; block += kBlockSize) {
-      memcpy(expected_bytes + block, in_bytes + block + 1, kBlockSize - 1);
-      expected_bytes[block + kBlockSize - 1] = 0;
-    }
-    HWY_ASSERT_VEC_EQ(d, expected.get(), ShiftRightBytes<1>(d, v));
-#else
-    (void)d;
-#endif  // #if HWY_TARGET != HWY_SCALAR
-  }
-};
-
-HWY_NOINLINE void TestAllShiftBytes() {
-  ForIntegerTypes(ForPartialVectors<TestShiftBytes>());
-}
-
-struct TestShiftLanes {
-  template <class T, class D>
-  HWY_NOINLINE void operator()(T /*unused*/, D d) {
-    // Scalar does not define Shift*Lanes.
-#if HWY_TARGET != HWY_SCALAR || HWY_IDE
-    const auto v = Iota(d, T(1));
-    const size_t N = Lanes(d);
-    auto expected = AllocateAligned<T>(N);
-
-    HWY_ASSERT_VEC_EQ(d, v, ShiftLeftLanes<0>(v));
-    HWY_ASSERT_VEC_EQ(d, v, ShiftLeftLanes<0>(d, v));
-    HWY_ASSERT_VEC_EQ(d, v, ShiftRightLanes<0>(d, v));
-
-    constexpr size_t kLanesPerBlock = 16 / sizeof(T);
-
-    for (size_t i = 0; i < N; ++i) {
-      expected[i] = (i % kLanesPerBlock) == 0 ? T(0) : T(i);
-    }
-    HWY_ASSERT_VEC_EQ(d, expected.get(), ShiftLeftLanes<1>(v));
-    HWY_ASSERT_VEC_EQ(d, expected.get(), ShiftLeftLanes<1>(d, v));
-
-    for (size_t i = 0; i < N; ++i) {
-      const size_t mod = i % kLanesPerBlock;
-      expected[i] = mod == (kLanesPerBlock - 1) || i >= N - 1 ? T(0) : T(2 + i);
-    }
-    HWY_ASSERT_VEC_EQ(d, expected.get(), ShiftRightLanes<1>(d, v));
-#else
-    (void)d;
-#endif  // #if HWY_TARGET != HWY_SCALAR
-  }
-};
-
-HWY_NOINLINE void TestAllShiftLanes() {
-  ForAllTypes(ForPartialVectors<TestShiftLanes>());
-}
-
 template <typename D, int kLane>
 struct TestBroadcastR {
   HWY_NOINLINE void operator()() const {
@@ -185,12 +91,12 @@ struct TestTableLookupBytes {
   HWY_NOINLINE void operator()(T /*unused*/, D d) {
 #if HWY_TARGET != HWY_SCALAR
     RandomState rng;
+
     const typename ChooseTableSize<kFull>::template type<T, D> d_tbl;
     const Repartition<uint8_t, decltype(d_tbl)> d_tbl8;
     const size_t NT8 = Lanes(d_tbl8);
 
     const Repartition<uint8_t, D> d8;
-    const size_t N = Lanes(d);
     const size_t N8 = Lanes(d8);
 
     // Random input bytes
@@ -208,7 +114,7 @@ struct TestTableLookupBytes {
         11, 10, 3, 4, 5,  8,  7,  6,  14, 13, 12, 15, 2,  1,  2,  0,
         4,  3,  2, 2, 5,  6,  7,  7,  15, 15, 15, 15, 15, 15, 0,  1};
     auto index_bytes = AllocateAligned<uint8_t>(N8);
-    const size_t max_index = HWY_MIN(N8, 16) - 1;
+    const size_t max_index = HWY_MIN(NT8, 16) - 1;
     for (size_t i = 0; i < N8; ++i) {
       index_bytes[i] = (i < 64) ? index_bytes_source[i] : 0;
       // Avoid asan error for partial vectors.
@@ -216,15 +122,21 @@ struct TestTableLookupBytes {
     }
     const auto indices = Load(d, reinterpret_cast<const T*>(index_bytes.get()));
 
+    const size_t N = Lanes(d);
     auto expected = AllocateAligned<T>(N);
     uint8_t* expected_bytes = reinterpret_cast<uint8_t*>(expected.get());
 
     for (size_t block = 0; block < N8; block += 16) {
       for (size_t i = 0; i < 16 && (block + i) < N8; ++i) {
         const uint8_t index = index_bytes[block + i];
-        HWY_ASSERT(block + index < N8);  // indices were already capped to N8.
-        // For large vectors, the lane index may wrap around due to block.
-        expected_bytes[block + i] = in_bytes[(block & 0xFF) + index];
+        HWY_ASSERT(index <= max_index);
+        // Note that block + index may exceed NT8 on RVV, which is fine because
+        // the operation uses the larger of the table and index vector size.
+        HWY_ASSERT(block + index < HWY_MAX(N8, NT8));
+        // For large vectors, the lane index may wrap around due to block,
+        // also wrap around after 8-bit overflow.
+        expected_bytes[block + i] =
+            in_bytes[(block + index) % HWY_MIN(NT8, 256)];
       }
     }
     HWY_ASSERT_VEC_EQ(d, expected.get(), TableLookupBytes(in, indices));
@@ -251,15 +163,14 @@ struct TestTableLookupBytes {
   }
 };
 
-HWY_NOINLINE void TestAllTableLookupBytes() {
+HWY_NOINLINE void TestAllTableLookupBytesSame() {
   // Partial index, same-sized table.
   ForIntegerTypes(ForPartialVectors<TestTableLookupBytes<false>>());
+}
 
-// TODO(janwas): requires LMUL trunc/ext, which is not yet implemented.
-#if HWY_TARGET != HWY_RVV
+HWY_NOINLINE void TestAllTableLookupBytesMixed() {
   // Partial index, full-size table.
   ForIntegerTypes(ForPartialVectors<TestTableLookupBytes<true>>());
-#endif
 }
 
 struct TestInterleaveLower {
@@ -327,28 +238,29 @@ struct TestZipLower {
     const size_t N = Lanes(d);
     auto even_lanes = AllocateAligned<T>(N);
     auto odd_lanes = AllocateAligned<T>(N);
+    // At least 2 lanes for HWY_SCALAR
+    auto zip_lanes = AllocateAligned<T>(HWY_MAX(N, 2));
+    const T kMaxT = LimitsMax<T>();
     for (size_t i = 0; i < N; ++i) {
-      even_lanes[i] = static_cast<T>(2 * i + 0);
-      odd_lanes[i] = static_cast<T>(2 * i + 1);
+      even_lanes[i] = static_cast<T>((2 * i + 0) & kMaxT);
+      odd_lanes[i] = static_cast<T>((2 * i + 1) & kMaxT);
     }
     const auto even = Load(d, even_lanes.get());
     const auto odd = Load(d, odd_lanes.get());
 
-    const Repartition<WideT, D> dw;
-    const size_t NW = Lanes(dw);
-    auto expected = AllocateAligned<WideT>(NW);
-    const size_t blockN = HWY_MIN(size_t(16) / sizeof(WideT), NW);
+    const size_t blockN = HWY_MIN(size_t(16) / sizeof(T), N);
 
-    for (size_t i = 0; i < NW; ++i) {
-      const size_t block = i / blockN;
-      // Value of least-significant lane in lo-vector.
-      const size_t lo = 2u * (i % blockN) + 4u * block * blockN;
-      const size_t kBits = sizeof(T) * 8;
-      expected[i] = static_cast<WideT>((static_cast<WideT>(lo + 1) << kBits) +
-                                       static_cast<WideT>(lo));
+    for (size_t i = 0; i < N; i += 2) {
+      const size_t base = (i / blockN) * blockN;
+      const size_t mod = i % blockN;
+      zip_lanes[i + 0] = even_lanes[mod / 2 + base];
+      zip_lanes[i + 1] = odd_lanes[mod / 2 + base];
     }
-    HWY_ASSERT_VEC_EQ(dw, expected.get(), ZipLower(even, odd));
-    HWY_ASSERT_VEC_EQ(dw, expected.get(), ZipLower(dw, even, odd));
+    const Repartition<WideT, D> dw;
+    const auto expected =
+        Load(dw, reinterpret_cast<const WideT*>(zip_lanes.get()));
+    HWY_ASSERT_VEC_EQ(dw, expected, ZipLower(even, odd));
+    HWY_ASSERT_VEC_EQ(dw, expected, ZipLower(dw, even, odd));
   }
 };
 
@@ -362,63 +274,54 @@ struct TestZipUpper {
     if (N < 16 / sizeof(T)) return;
     auto even_lanes = AllocateAligned<T>(N);
     auto odd_lanes = AllocateAligned<T>(N);
-    for (size_t i = 0; i < Lanes(d); ++i) {
-      even_lanes[i] = static_cast<T>(2 * i + 0);
-      odd_lanes[i] = static_cast<T>(2 * i + 1);
+    auto zip_lanes = AllocateAligned<T>(N);
+    const T kMaxT = LimitsMax<T>();
+    for (size_t i = 0; i < N; ++i) {
+      even_lanes[i] = static_cast<T>((2 * i + 0) & kMaxT);
+      odd_lanes[i] = static_cast<T>((2 * i + 1) & kMaxT);
     }
     const auto even = Load(d, even_lanes.get());
     const auto odd = Load(d, odd_lanes.get());
 
-        const Repartition<WideT, D> dw;
-    const size_t NW = Lanes(dw);
-    auto expected = AllocateAligned<WideT>(NW);
-    const size_t blockN = HWY_MIN(size_t(16) / sizeof(WideT), NW);
+    const size_t blockN = HWY_MIN(size_t(16) / sizeof(T), N);
 
-    for (size_t i = 0; i < NW; ++i) {
-      const size_t block = i / blockN;
-      const size_t lo = 2u * (i % blockN) + 4u * block * blockN;
-      const size_t kBits = sizeof(T) * 8;
-      expected[i] = static_cast<WideT>(
-          (static_cast<WideT>(lo + 2 * blockN + 1) << kBits) +
-          static_cast<WideT>(lo + 2 * blockN));
+    for (size_t i = 0; i < N; i += 2) {
+      const size_t base = (i / blockN) * blockN + blockN / 2;
+      const size_t mod = i % blockN;
+      zip_lanes[i + 0] = even_lanes[mod / 2 + base];
+      zip_lanes[i + 1] = odd_lanes[mod / 2 + base];
     }
-    HWY_ASSERT_VEC_EQ(dw, expected.get(), ZipUpper(dw, even, odd));
+    const Repartition<WideT, D> dw;
+    const auto expected =
+        Load(dw, reinterpret_cast<const WideT*>(zip_lanes.get()));
+    HWY_ASSERT_VEC_EQ(dw, expected, ZipUpper(dw, even, odd));
   }
 };
 
 HWY_NOINLINE void TestAllZip() {
   const ForDemoteVectors<TestZipLower> lower_unsigned;
-  // TODO(janwas): enable after LowerHalf available
-#if HWY_TARGET != HWY_RVV
   lower_unsigned(uint8_t());
-#endif
   lower_unsigned(uint16_t());
 #if HWY_HAVE_INTEGER64
   lower_unsigned(uint32_t());  // generates u64
 #endif
 
   const ForDemoteVectors<TestZipLower> lower_signed;
-#if HWY_TARGET != HWY_RVV
   lower_signed(int8_t());
-#endif
   lower_signed(int16_t());
 #if HWY_HAVE_INTEGER64
   lower_signed(int32_t());  // generates i64
 #endif
 
   const ForShrinkableVectors<TestZipUpper> upper_unsigned;
-#if HWY_TARGET != HWY_RVV
   upper_unsigned(uint8_t());
-#endif
   upper_unsigned(uint16_t());
 #if HWY_HAVE_INTEGER64
   upper_unsigned(uint32_t());  // generates u64
 #endif
 
   const ForShrinkableVectors<TestZipUpper> upper_signed;
-#if HWY_TARGET != HWY_RVV
   upper_signed(int8_t());
-#endif
   upper_signed(int16_t());
 #if HWY_HAVE_INTEGER64
   upper_signed(int32_t());  // generates i64
@@ -427,108 +330,6 @@ HWY_NOINLINE void TestAllZip() {
   // No float - concatenating f32 does not result in a f64
 }
 
-// Scalar does not define CombineShiftRightBytes.
-#if HWY_TARGET != HWY_SCALAR || HWY_IDE
-
-template <int kBytes>
-struct TestCombineShiftRightBytes {
-  template <class T, class D>
-  HWY_NOINLINE void operator()(T, D d) {
-    const size_t kBlockSize = 16;
-    static_assert(kBytes < kBlockSize, "Shift count is per block");
-    const Repartition<uint8_t, D> d8;
-    const size_t N8 = Lanes(d8);
-    if (N8 < 16) return;
-    auto hi_bytes = AllocateAligned<uint8_t>(N8);
-    auto lo_bytes = AllocateAligned<uint8_t>(N8);
-    auto expected_bytes = AllocateAligned<uint8_t>(N8);
-    uint8_t combined[2 * kBlockSize];
-
-    // Random inputs in each lane
-    RandomState rng;
-    for (size_t rep = 0; rep < AdjustedReps(100); ++rep) {
-      for (size_t i = 0; i < N8; ++i) {
-        hi_bytes[i] = static_cast<uint8_t>(Random64(&rng) & 0xFF);
-        lo_bytes[i] = static_cast<uint8_t>(Random64(&rng) & 0xFF);
-      }
-      for (size_t i = 0; i < N8; i += kBlockSize) {
-        CopyBytes<kBlockSize>(&lo_bytes[i], combined);
-        CopyBytes<kBlockSize>(&hi_bytes[i], combined + kBlockSize);
-        CopyBytes<kBlockSize>(combined + kBytes, &expected_bytes[i]);
-      }
-
-      const auto hi = BitCast(d, Load(d8, hi_bytes.get()));
-      const auto lo = BitCast(d, Load(d8, lo_bytes.get()));
-      const auto expected = BitCast(d, Load(d8, expected_bytes.get()));
-      HWY_ASSERT_VEC_EQ(d, expected, CombineShiftRightBytes<kBytes>(d, hi, lo));
-    }
-  }
-};
-
-template <int kLanes>
-struct TestCombineShiftRightLanes {
-  template <class T, class D>
-  HWY_NOINLINE void operator()(T, D d) {
-    const Repartition<uint8_t, D> d8;
-    const size_t N8 = Lanes(d8);
-    if (N8 < 16) return;
-
-    auto hi_bytes = AllocateAligned<uint8_t>(N8);
-    auto lo_bytes = AllocateAligned<uint8_t>(N8);
-    auto expected_bytes = AllocateAligned<uint8_t>(N8);
-    const size_t kBlockSize = 16;
-    uint8_t combined[2 * kBlockSize];
-
-    // Random inputs in each lane
-    RandomState rng;
-    for (size_t rep = 0; rep < AdjustedReps(100); ++rep) {
-      for (size_t i = 0; i < N8; ++i) {
-        hi_bytes[i] = static_cast<uint8_t>(Random64(&rng) & 0xFF);
-        lo_bytes[i] = static_cast<uint8_t>(Random64(&rng) & 0xFF);
-      }
-      for (size_t i = 0; i < N8; i += kBlockSize) {
-        CopyBytes<kBlockSize>(&lo_bytes[i], combined);
-        CopyBytes<kBlockSize>(&hi_bytes[i], combined + kBlockSize);
-        CopyBytes<kBlockSize>(combined + kLanes * sizeof(T),
-                              &expected_bytes[i]);
-      }
-
-      const auto hi = BitCast(d, Load(d8, hi_bytes.get()));
-      const auto lo = BitCast(d, Load(d8, lo_bytes.get()));
-      const auto expected = BitCast(d, Load(d8, expected_bytes.get()));
-      HWY_ASSERT_VEC_EQ(d, expected, CombineShiftRightLanes<kLanes>(d, hi, lo));
-    }
-  }
-};
-
-#endif  // #if HWY_TARGET != HWY_SCALAR
-
-struct TestCombineShiftRight {
-  template <class T, class D>
-  HWY_NOINLINE void operator()(T t, D d) {
-// Scalar does not define CombineShiftRightBytes.
-#if HWY_TARGET != HWY_SCALAR || HWY_IDE
-    constexpr int kMaxBytes = HWY_MIN(16, int(MaxLanes(d) * sizeof(T)));
-    constexpr int kMaxLanes = kMaxBytes / static_cast<int>(sizeof(T));
-    TestCombineShiftRightBytes<kMaxBytes - 1>()(t, d);
-    TestCombineShiftRightBytes<HWY_MAX(kMaxBytes / 2, 1)>()(t, d);
-    TestCombineShiftRightBytes<1>()(t, d);
-
-    TestCombineShiftRightLanes<kMaxLanes - 1>()(t, d);
-    TestCombineShiftRightLanes<HWY_MAX(kMaxLanes / 2, -1)>()(t, d);
-    TestCombineShiftRightLanes<1>()(t, d);
-#else
-    (void)t;
-    (void)d;
-#endif
-  }
-};
-
-HWY_NOINLINE void TestAllCombineShiftRight() {
-  // Need at least 2 lanes.
-  ForAllTypes(ForShrinkableVectors<TestCombineShiftRight>());
-}
-
 class TestSpecialShuffle32 {
  public:
   template <class T, class D>
@@ -619,20 +420,12 @@ HWY_AFTER_NAMESPACE();
 
 namespace hwy {
 HWY_BEFORE_TEST(HwyBlockwiseTest);
-HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllShiftBytes);
-HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllShiftLanes);
 HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllBroadcast);
-HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllTableLookupBytes);
+HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllTableLookupBytesSame);
+HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllTableLookupBytesMixed);
 HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllInterleave);
 HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllZip);
-HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllCombineShiftRight);
 HWY_EXPORT_AND_TEST_P(HwyBlockwiseTest, TestAllSpecialShuffles);
 }  // namespace hwy
 
-// Ought not to be necessary, but without this, no tests run on RVV.
-int main(int argc, char** argv) {
-  ::testing::InitGoogleTest(&argc, argv);
-  return RUN_ALL_TESTS();
-}
-
 #endif
